import ast
import base64
import datetime
import msvcrt
import re
import subprocess
import sys
from argparse import ArgumentParser
from pathlib import Path

import ast_comments as ast
import colorama
import compyner.__main__
import compyner.engine
import serial.tools.list_ports_windows
import tqdm

from . import spikeapi
from .spike_prime_compyne import spike_prime_compyne


def greet():
    print(colorama.Fore.BLUE + compyner.__main__.ASCII_LOGO + colorama.Style.RESET_ALL)
    print(f"{'for Spike Prime':>48}")
    print()


def get_device():
    print(colorama.Fore.GREEN + "> Searching for devices..." + colorama.Style.RESET_ALL)
    devices = serial.tools.list_ports_windows.comports()
    if len(devices) == 0:
        print(
            colorama.Fore.RED
            + colorama.Style.BRIGHT
            + "Error: No devices found"
            + colorama.Style.RESET_ALL
        )
        sys.exit(1)
    if len(devices) == 1:
        device_choice = devices[0]
    else:
        for index, device in enumerate(devices):
            print(f"{index+1:>2}. {device.device}")
        device_choice = devices[int(input("Device: ")) - 1]

    print(colorama.Fore.GREEN + "> Connecting..." + colorama.Fore.RESET)
    return spikeapi.Device(device_choice.device)


def show_stdout(device: spikeapi.Device, error_replace_location: Path | None):
    if error_replace_location:
        lineno_map = compyner.engine.get_lineno_map(
            ast.parse(error_replace_location.read_text("utf-8"))
        )

        def error_replacer(match: re.Match[str]):
            match_str = match.group(0)
            if not match_str.startswith(f'"{error_replace_location}"'):
                return match_str
            mapped = lineno_map.get(int(match_str.rsplit("line ", 1)[1][:-1]))
            if mapped is None:
                return match_str
            return f'(comPYned) "{mapped}"'

    print(colorama.Fore.CYAN + ">> Press any key to exit" + colorama.Fore.RESET)
    while True:
        if not device.active:
            print(colorama.Fore.RED + "> Device disconnected" + colorama.Fore.RESET)
            break
        if device.logs:
            log = device.logs.popleft()
            if log.type == spikeapi.LogType.PRINT:
                print(colorama.Fore.LIGHTBLACK_EX + log.entry + colorama.Fore.RESET)
            if log.type == spikeapi.LogType.USER_PROGRAM_PRINT:
                print(log.entry)
            elif log.type == spikeapi.LogType.USER_PROGRAM_ERROR:
                if error_replace_location:
                    log.entry = re.sub(r'".*", line \d*,', error_replacer, log.entry)
                print(colorama.Fore.RED + log.entry + colorama.Fore.RESET)
            elif log.type == spikeapi.LogType.RUNTIME_ERROR:
                print(colorama.Fore.YELLOW + log.entry + colorama.Fore.RESET)

        if msvcrt.kbhit():
            print(
                colorama.Fore.LIGHTGREEN_EX
                + "> Got input. Exiting..."
                + colorama.Fore.RESET
            )
            break


def wipe_slots(device: spikeapi.Device, slots: list[int]):
    print(colorama.Fore.GREEN + "> Wiping slots..." + colorama.Fore.RESET)
    existing_slots = device.get_storage_information().wait_for_response().slots.keys()
    for slot in slots:
        if str(slot) in existing_slots:
            device.wipe_slot(slot).wait_for_response()

def move_slots(device: spikeapi.Device, from_slot:int, to_slot: int):
    print(colorama.Fore.GREEN + "> Moving slot..." + colorama.Fore.RESET)
    device.move_slot(from_slot, to_slot).wait_for_response()


def start(
    device: spikeapi.Device,
    slot: int = 0,
    file: Path | None = None,
    hide_output: bool = False,
    wait: bool = True
):
    print(colorama.Fore.GREEN + "> Running..." + colorama.Fore.RESET)
    device.run_program(slot)

    if wait:
        while not device.running_program:
            pass

    if not hide_output:
        show_stdout(device, file)


def stop(
    device: spikeapi.Device,
    wait: bool = True
):
    print(colorama.Fore.GREEN + "> Terminating..." + colorama.Fore.RESET)
    device.terminate_program()

    if wait:
        while device.running_program:
            pass
        


def upload(
    device: spikeapi.Device,
    input: Path,
    slot: int = 0,
    debug: bool = False,
    autostart: bool = True,
    hide_output: bool = False,
):
    # Step 1: ComPYning
    print(colorama.Fore.GREEN + "> ComPYning..." + colorama.Fore.RESET)
    comPYned: str = spike_prime_compyne(input, slot=slot, debug_build=debug)
    input.with_suffix(".cpyd.py").write_text(comPYned, "utf-8")

    # Step 2: Compiling
    print(colorama.Fore.GREEN + "> Compiling..." + colorama.Fore.RESET, end="")
    proc = subprocess.run(
        ["mpy-cross-v5", input.with_suffix(".cpyd.py").absolute()],
        check=False,
        stderr=subprocess.PIPE,
    )
    if proc.returncode != 0:
        print(
            colorama.Style.BRIGHT
            + colorama.Fore.RED
            + "\nFailed:"
            + colorama.Style.NORMAL
        )
        print(proc.stderr.decode("utf-8") + colorama.Style.RESET_ALL)
        return
    mpy = input.with_suffix(".cpyd.mpy").absolute().read_bytes()
    input.with_suffix(".cpyd.mpy").absolute().unlink()
    print(colorama.Fore.GREEN + " done" + colorama.Fore.RESET)

    # Step 3: Uploading
    print(colorama.Fore.GREEN + "> Uploading..." + colorama.Fore.RESET)
    progress_bar = tqdm.tqdm(total=len(mpy), unit="B", unit_scale=True)

    def callback(done, total, bs):
        progress_bar.update(bs)

    device.upload_file(
        mpy,
        slot,
        input.name,
        filename="__init__.mpy",
        callback=callback,
    )

    progress_bar.close()

    if autostart:
        start(device, slot, input.with_suffix(".cpyd.py"), hide_output)


def show_slots(device: spikeapi.Device):
    print(colorama.Fore.GREEN + "> Loading data..." + colorama.Fore.RESET)
    data = device.get_storage_information().wait_for_response()
    print("Slots:")
    for slot_id, slot in data.slots.items():
        print(f"    {slot_id:>2}:")
        print(f"        ID          : {slot.id}")
        print(f"        Name        : {base64.b64decode(slot.name).decode("utf-8")}")
        print(f"        Project ID  : {slot.project_id}")
        print(f"        Type        : {slot.type}")
        print(f"        Size        : {slot.size} kb")
        print(
            f"        Created At  : {datetime.datetime.fromtimestamp(slot.created / 1000).strftime("%d.%m.%Y %H:%M:%S")}"
        )
        print(
            f"        Modified At : {datetime.datetime.fromtimestamp(slot.modified / 1000).strftime("%d.%m.%Y %H:%M:%S")}"
        )
    if not data.slots:
        print("    No slots.")
    print()


def show_storage(device: spikeapi.Device):
    print(colorama.Fore.GREEN + "> Loading data..." + colorama.Fore.RESET)
    data = device.get_storage_information().wait_for_response()
    print("Storage:")
    print(f"    Availible : {data.storage.available:>8} {data.storage.unit}")
    print(f"    Free      : {data.storage.free:>8} {data.storage.unit}")
    print(f"    Total     : {data.storage.total:>8} {data.storage.unit}")
    print(f"    Used      : {data.storage.pct:>8} %")
    print()


def show_firmware(device: spikeapi.Device):
    print(colorama.Fore.GREEN + "> Loading data..." + colorama.Fore.RESET)
    data = device.get_firmware_info().wait_for_response()
    print("Firmware:")
    print(f"    Checksum         : {data.firmware.checksum}")
    print(f"    Firmware Version : {".".join(map(str, data.firmware.version))}")
    print(f"    Runtime Version  : {".".join(map(str, data.runtime.version))}")
    print(f"    Capabilities     : {", ".join(data.capabilities)}")
    print(f"    Variant          : {data.variant}")
    print(f"    USB PID          : {data.usb_pid}")
    print()


def show_uuid(device: spikeapi.Device):
    print(colorama.Fore.GREEN + "> Loading data..." + colorama.Fore.RESET)
    data = device.get_firmware_info().wait_for_response()
    print(f"Device UUID: {data.device_uuid}")
    
def show_battery(device: spikeapi.Device):
    print(colorama.Fore.GREEN + "> Loading data..." + colorama.Fore.RESET)
    device.start_repl()
    device.exec_in_repl("import hub")
    charger_detect = device.eval_in_repl("hub.battery.charger_detect()")
    voltage = device.eval_in_repl("hub.battery.voltage()")
    current = device.eval_in_repl("hub.battery.current()")
    capacity_left = device.eval_in_repl("hub.battery.capacity_left()")
    temperature = device.eval_in_repl("hub.battery.temperature()")
    print("Battery:")
    print(f"    Capacity Left : {capacity_left} %")
    print(f"    Charging      : {charger_detect}")
    print(f"    Voltage       : {voltage} mV")
    print(f"    Current       : {current} ?")
    print(f"    Temperature   : {temperature} °C")
    print()
    device.soft_reboot()


def main() -> None:
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest="action")
    upload_parser = subparsers.add_parser("upload")
    upload_parser.add_argument(
        "input",
        action="store",
        type=compyner.__main__.file_path_exists,
    )
    upload_parser.add_argument(
        "--slot",
        "-s",
        required=False,
        action="store",
        type=int,
        default=0,
    )
    upload_parser.add_argument("--debug", action="store_true")
    upload_parser.add_argument("--start", action="store_true")
    upload_parser.add_argument(
        "--read",
        action="store_true",
        help="show program output",
    )
    read_parser = subparsers.add_parser("read")
    read_parser.add_argument(
        "--file",
        action="store",
        type=compyner.__main__.file_path_exists,
        required=False,
        default=None,
        help="path to a compyned file. Will be used to replace error locations",
    )
    start_parser = subparsers.add_parser("start")
    start_parser.add_argument("slot", action="store", type=int)
    start_parser.add_argument(
        "--read",
        action="store_true",
        help="dont show program output",
    )
    start_parser.add_argument(
        "--file",
        action="store",
        type=compyner.__main__.file_path_exists,
        required=False,
        default=None,
        help="path to a compyned file. Will be used to replace error locations",
    )
    start_parser.add_argument(
        "--wait",
        action="store_true",
        help="wait until program is started",
    )
    stop_parser = subparsers.add_parser("stop")
    stop_parser.add_argument(
        "--wait",
        action="store_true",
        help="wait until program is stopped",
    )
    subparsers.add_parser("poweroff")
    reboot_parser = subparsers.add_parser("reboot")
    reboot_parser.add_argument(
        "--hard",
        action="store_true",
        help="fully reboot",
    )
    wipe_parser = subparsers.add_parser("wipe")
    opt_group = wipe_parser.add_mutually_exclusive_group(required=True)
    opt_group.add_argument("slot", action="store", nargs="*", type=int)
    opt_group.add_argument("--all", action="store_true")
    move_parser = subparsers.add_parser("move")
    move_parser.add_argument("from_slot", action="store", type=int)
    move_parser.add_argument("to_slot", action="store", type=int)
    get_parser = subparsers.add_parser("get")
    get_parser.add_argument("detail", choices=["storage", "slots", "firmware", "uuid", "battery"])
    args = parser.parse_args()

    try:
        if args.action == "upload":
            device = get_device()
            upload(
                device,
                args.input,
                args.slot,
                args.debug,
                args.start,
                not args.read,
            )
        elif args.action == "start":
            device = get_device()
            start(device, args.slot, args.file, not args.read, args.wait)
        elif args.action == "stop":
            device = get_device()
            stop(device, args.wait)
        elif args.action == "read":
            device = get_device()
            show_stdout(device, args.file)
        elif args.action == "wipe":
            device = get_device()
            wipe_slots(device, range(16) if args.all else args.slot)
        elif args.action == "move":
            device = get_device()
            move_slots(device, args.from_slot, args.to_slot)
        elif args.action == "get":
            device = get_device()
            if args.detail == "storage":
                show_storage(device)
            elif args.detail == "slots":
                show_slots(device)
            elif args.detail == "uuid":
                show_uuid(device)
            elif args.detail == "battery":
                show_battery(device)
            else:
                show_firmware(device)
        elif args.action == "poweroff":
            device = get_device()
            device.start_repl()
            device.exec_in_repl("import hub")
            device.exec_in_repl("hub.power_off(fast=True)")
        elif args.action == "reboot":
            device = get_device()
            if args.hard:
                device.start_repl()
                device.exec_in_repl("import hub")
                device.exec_in_repl("hub.power_off(restart=True)")
            else:
                device.start_repl()
                device.soft_reboot()
        else:
            parser.error("Invalid action")
    except ConnectionError as e:
        print("Lost connection", e)


if __name__ == "__main__":
    main()



import base64
from collections import deque
import enum
import json
import math
import random
import string
import threading
import time
from typing import Any, Literal, Optional
import serial
import serial.serialutil
import serial.tools.list_ports_windows as serial_tool_search
from dataclasses import dataclass
from dataclasses_json import DataClassJsonMixin

letters = string.ascii_letters + string.digits + "_"


def get_arrow(done, todo, blocksize):
    blocks = math.ceil(todo / blocksize)
    done_blocks = math.ceil(done / blocksize)
    done_perquartercent = done_blocks / blocks * 25

    def h(x):
        return 9 if done_perquartercent > x else 6

    def l(x):
        return 4 if done_perquartercent > x else 0

    img = (
        (l, l, h, l, l),
        (l, l, h, l, l),
        (h, h, h, h, h),
        (l, h, h, h, l),
        (l, l, h, l, l),
    )
    result = tuple(
        tuple(pixel(y * 5 + x) for x, pixel in enumerate(row))
        for y, row in enumerate(img)
    )
    return ":".join(["".join([str(pixel) for pixel in row]) for row in result])


def random_id(len=4):
    return "".join(random.choice(letters) for _ in range(len))


class Gesture(enum.Enum):
    TAPPED = 0
    DOUBLETAPPED = 1
    SHAKEN = 2
    FALLING = 3

    def __json__(self):
        return [self.value, self.name]


class Orientation(enum.Enum):
    BATTERY_DOWN = 0
    SPEAKER_DOWN = 1
    LEFT_DOWN = 2
    SCREEN_DOWN = 3
    PORT_DOWN = 4
    RIGHT_DOWN = 5

    def __json__(self):
        return [self.value, self.name]


@dataclass
class JsonMessage(DataClassJsonMixin):
    m: Optional[int | str] = None
    "The message type. Not there if response."
    i: Optional[str] = None
    "A four letter ID to make the massage unique and answerable"
    p: Optional[Any] = None
    "Additional information. Depending on message type"
    r: Optional[Any] = None
    "Result. Only if message is a response. Depending on message type"
    e: Optional[Any] = None
    "Error. Only if message is a response. Depending on message type"


@dataclass
class JsonFirmwareInfoRuntime(DataClassJsonMixin):
    version: list[int]


@dataclass
class JsonFirmwareInfoFirmware(DataClassJsonMixin):
    checksum: str
    version: list[int]


@dataclass
class JsonFirmwareInfo(DataClassJsonMixin):
    runtime: JsonFirmwareInfoRuntime
    capabilities: list[str]
    firmware: JsonFirmwareInfoFirmware
    usb_pid: str
    variant: str
    device_uuid: str


@dataclass
class JsonStorageInfoSlot(DataClassJsonMixin):
    id: int
    size: int
    "size in kb"
    name: str
    "Project name, set on upload"
    project_id: str
    "Project ID, set on upload"
    modified: int
    "Modified unix timestamp"
    created: int
    "Created unix timestamp"
    type: Literal["python"] | Literal["scratch"]
    "Projects with python have the normal python API, while projects with scratch have a more advanced API and are async"


@dataclass
class JsonStorageInfoStorage(DataClassJsonMixin):
    available: int
    "Available storage in given .unit"
    total: int
    "Total storage in given .unit"
    pct: float
    "percentage free"
    unit: str
    "unit for storage, seems to be kb always"
    free: int
    "Free storage in given .unit"


@dataclass
class JsonStorageInfo(DataClassJsonMixin):
    storage: JsonStorageInfoStorage
    slots: dict[str, JsonStorageInfoSlot]
    "Dict of str-ints and slots. str-int: '9' instead of 9. why? idk"


class Message[T]:
    there: bool
    id: str
    result: Optional[T]
    error: Optional[str]

    def __init__(
        self, device: "Device", id: int, response_parser: DataClassJsonMixin = None
    ):
        self.device = device
        self.id = id

        self._response_parser = response_parser

        self.device._sent_messages[self.id] = self

        self.there = False
        self.result = None
        self.error = None

    def wait_for_response(self) -> T:
        while not self.there:
            if not self.device.active:
                raise ConnectionAbortedError("Device disconnected.")
        if self.error:
            error = self.error
            try:
                error = base64.b64decode(error)
            except:
                pass
            try:
                error = json.loads(error)
            except:
                pass
            raise ConnectionError(error)
        return (
            self._response_parser(self.result) if self._response_parser else self.result
        )


class LogType(enum.Enum):
    USER_PROGRAM_PRINT = enum.auto()
    USER_PROGRAM_ERROR = enum.auto()
    RUNTIME_ERROR = enum.auto()
    PRINT = enum.auto()


class LogEntry:
    def __init__(self, type: LogType, entry: str):
        self.type = type
        self.entry = entry

    def __str__(self):
        return f"{self.type}: {self.entry}"

    def __repr__(self):
        return f"<LogEntry {str(self)}>"


class Device:
    def __init__(self, port: str):
        self.port = port
        self._ser = serial.Serial(port, 115200)
        self._sent_messages: dict[str, Message] = {}
        self.logs: deque[LogEntry] = deque()
        self._current_image = None
        self._buffer = bytearray()
        self._ran_first = False
        self._repl_prompt = False
        self._repl_in = []
        self._repl_out = []
        self.active = True

        self._tickloop_thread = threading.Thread(target=self.tickloop, daemon=True)
        self._tickloop_thread.start()

        self.running_program: bool = False
        self.orientation: Optional[Orientation] = None
        self.last_gesture: Optional[Gesture] = None

    def ensure_connected(self):
        if not self.active:
            raise ConnectionError("Device not connected anymore.")

    def send_message(self, name, params=None, response_parser=None) -> Message:
        self.ensure_connected()
        while not self._ran_first:
            ...
        id = random_id()
        message = Message(self, id, response_parser)
        msg = {"m": name, "p": params, "i": id}
        msg_string = json.dumps(msg)
        self._ser.write(msg_string.encode("utf-8"))
        self._ser.write(b"\x0d")
        return message

    def handle_message(self, rmsg: bytes):
        if not rmsg:
            return
        self._ran_first = True
        if rmsg.startswith(b"\n<<< "):
            self._repl_out.append(rmsg.decode("utf-8")[5:])
        elif rmsg.startswith(b"\n>>> "):
            self._repl_in.append(rmsg.decode("utf-8")[5:])
        try:
            msg = JsonMessage.from_json(rmsg.decode("utf-8"))
        except (json.JSONDecodeError, AttributeError):
            try:
                self.logs.append(LogEntry(LogType.PRINT, rmsg.decode("utf-8")))
            except UnicodeDecodeError:
                self.logs.append(LogEntry(LogType.PRINT, "?"))
            return
        mid = msg.i
        message = msg.m

        if mid and mid in self._sent_messages:
            pending_message = self._sent_messages.pop(mid)
            pending_message.result = msg.r
            pending_message.error = msg.e
            pending_message.there = True
        else:
            if message == "userProgram.print":
                self.logs.append(
                    LogEntry(LogType.USER_PROGRAM_PRINT, msg.p.get("value"))
                )
            elif message == "user_program_error":
                try:
                    decoded_line_1 = base64.b64decode(msg.p[3].encode("utf-8")).decode(
                        "utf-8"
                    )
                except (UnicodeDecodeError, UnicodeEncodeError):
                    decoded_line_1 = msg.p[3]
                try:
                    decoded_line_2 = base64.b64decode(msg.p[4].encode("utf-8")).decode(
                        "utf-8"
                    )
                except (UnicodeDecodeError, UnicodeEncodeError):
                    decoded_line_2 = msg.p[4]
                self.logs.append(
                    LogEntry(
                        LogType.USER_PROGRAM_ERROR,
                        decoded_line_1 + "\n" + decoded_line_2,
                    )
                )
            elif message == "runtime_error":
                self.logs.append(
                    LogEntry(
                        LogType.RUNTIME_ERROR,
                        base64.b64decode(msg.p[3].encode("utf-8")).decode("utf-8"),
                    )
                )
            elif message == 3:
                ...  # Button press button = p[0], time = p[1]. if time == 0, button down, else up.
            elif message == 4:
                self.last_gesture = Gesture(msg.p)
            elif message == 12:
                self.running_program = msg.p[1]
            elif message == 14:
                self.orientation = Orientation(msg.p)
            elif message in (0, 1, 2):
                ...  # print("?", msg)
            else:
                ...  # print("?", msg)

    def tick(self):
        self._buffer += self._ser.read_all()
        if self._buffer[-5:] == b"\n>>> ":
            self._repl_prompt = True
        else:
            self._repl_prompt = False
        pos = self._buffer.find(b"\x0d")
        if pos != -1:
            part = self._buffer[:pos]
            self.handle_message(part)
            self._buffer = self._buffer[pos + 1 :]

    def tickloop(self):
        while self.active:
            try:
                self.tick()
            except serial.serialutil.SerialException:
                self.active = False

    def display_set_pixel(self, x, y, brightness=9) -> Message[None]:
        self.ensure_connected()
        return self.send_message(
            "scratch.display_set_pixel", {"x": x, "y": y, "brightness": brightness}
        )

    def display_clear(self) -> Message[None]:
        self.ensure_connected()
        return self.send_message("scratch.display_clear")

    def _wait_for_prompt(self, timeout=3):
        start = time.time()
        while (time.time() - start) < timeout:
            if self._repl_prompt:
                return
        raise ConnectionError(
            "failed to get to the command prompt"
        )

    def exec_in_repl(self, cmd: str):
        self._ser.write(cmd.encode("utf-8") + b"\r\n")
        self._repl_prompt = False
        time.sleep(0.5)
        self._wait_for_prompt()

    def start_repl(self):
        self.ensure_connected()
        self._ser.write(b"\x03")
        self._ser.write(b"\x02")
        time.sleep(0.5)
        self._wait_for_prompt()

    def soft_reboot(self):
        self._ser.write(b"\04")
        time.sleep(0.5)
        
    def eval_in_repl(self, code: str):
        pos = len(self._repl_out)
        self.exec_in_repl("print('<<<', %s)" % code)
        while not len(self._repl_out) - pos:
            ...
        return self._repl_out[-1]

    def wipe_slot(self, slot: int) -> Message[None]:
        return self.send_message("remove_project", {"slotid": slot})

    def move_slot(self, from_slot: int, to_slot: int) -> Message[None]:
        return self.send_message(
            "move_project", {"old_slotid": from_slot, "new_slotid": to_slot}
        )

    def display_image(self, image: str) -> Message[None]:
        self.ensure_connected()
        return self.send_message("scratch.display_image", {"image": image})

    def display_image_for(self, image: str, duration_ms: int) -> Message[None]:
        self.ensure_connected()
        return self.send_message(
            "scratch.display_image_for", {"image": image, "duration": duration_ms}
        )

    def display_text(self, text: str | int) -> Message[None]:
        self.ensure_connected()
        return self.send_message("scratch.display_text", {"text": text})

    def upload_file(
        self,
        data: bytes,
        to_slot: int,
        name=None,
        mode="python",
        callback=None,
        filename=None,
    ):
        self.ensure_connected()

        now = int(time.time() * 1000)
        size = len(data)

        # self.display_image("00000:00000:90909:00000:00000").wait_for_response()
        # self.update_display(get_arrow(0), refresh=True)

        start = self.send_message(
            "start_write_program",
            {
                "slotid": to_slot,
                "size": size,
                "filename": filename,
                "meta": {
                    "created": now,
                    "modified": now,
                    "name": base64.b64encode(
                        (name or random_id(10)).encode("utf-8")
                    ).decode("utf-8"),
                    "type": mode,
                    "project_id": random_id(12),
                },
            },
        ).wait_for_response()

        bs = start["blocksize"]
        tid = start["transferid"]
        transferred = 0

        # self.display_image(get_arrow(transferred, size, bs)).wait_for_response()

        b = data[:bs]
        data = data[bs:]
        while b:
            self.send_message(
                "write_package",
                {"data": str(base64.b64encode(b), "utf-8"), "transferid": tid},
            ).wait_for_response()
            transferred += len(b)
            callback(transferred, size, bs)
            # self.display_image(get_arrow(transferred, size, bs)).wait_for_response()
            b = data[:bs]
            data = data[bs:]
        # time.sleep(0.1)

    def run_program(self, slot: int) -> Message[None]:
        self.ensure_connected()
        return self.send_message("program_execute", {"slotid": slot})

    def terminate_program(self) -> Message[None]:
        self.ensure_connected()
        return self.send_message("program_terminate")

    def get_storage_information(self) -> Message[JsonStorageInfo]:
        self.ensure_connected()
        return self.send_message(
            "get_storage_status", response_parser=JsonStorageInfo.from_dict
        )

    def get_firmware_info(self) -> Message[JsonFirmwareInfo]:
        self.ensure_connected()
        return self.send_message(
            "get_hub_info", response_parser=JsonFirmwareInfo.from_dict
        )

    def __del__(self):
        self.active = False
        self._tickloop_thread.join()
        self._ser.close()


class DeviceManager:
    def __init__(self):
        self.connected_devices: dict[str, Device] = {}

        self.active = True
        self._tickloop_thread = threading.Thread(target=self.tickloop, daemon=True)
        self._tickloop_thread.start()

    def new_device(self, port: str):
        new_device = Device(port)
        self.connected_devices[port] = new_device

    def discover_comports(self):
        for port in serial_tool_search.comports():
            if port.name not in self.connected_devices:
                self.new_device(port.name)

    def remove_inactive_devices(self):
        self.connected_devices = {
            port: device
            for port, device in self.connected_devices.items()
            if device.active
        }

    def tick(self):
        self.remove_inactive_devices()
        self.discover_comports()

    def tickloop(self):
        while True:
            self.tick()
            time.sleep(1)

    def __del__(self):
        self.active = False
        self._tickloop_thread.join()



import ast_comments as ast
from pathlib import Path
import sys
from compyner.engine import ComPYner, ast_from_file


SPIKE_PRIME_MODULES = [
    "array",
    "binascii",
    "builtins",
    "cmath",
    "collections",
    "errno",
    "gc",
    "hashlib",
    "heapq",
    "io",
    "json",
    "math",
    "os",
    "random",
    "re",
    "select",
    "struct",
    "sys",
    "time",
    "zlib",
    "bluetooth",
    "machine",
    "micropython",
    "uctypes",
    "__main__",
    "_onewire",
    "firmware",
    "hub",
    "uarray",
    "ubinascii",
    "ubluetooth",
    "ucollections",
    "uerrno",
    "uhashlib",
    "uheapq",
    "uio",
    "ujson",
    "umachine",
    "uos",
    "urandom",
    "ure",
    "uselect",
    "utime",
    "utimeq",
    "uzlib",
    "spike",
    "mindstorms",
    "hub",
    "runtime",
]


class PreOptimize(ast.NodeTransformer):
    def visit_FunctionDef(self, node):
        for arg in node.args.args:
            arg.annotation = None
        for arg in node.args.posonlyargs:
            arg.annotation = None
        for arg in node.args.kwonlyargs:
            arg.annotation = None
        node.returns = None
        node = self.generic_visit(node)
        node.body = [
            stmt
            for stmt in node.body
            if not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant))
        ] or [ast.copy_location(ast.Pass(), node)]
        return node

    def visit_Comment(self, node):
        return None

    def visit_AnnAssign(self, node):
        if node.value:
            return self.generic_visit(
                ast.copy_location(
                    ast.Assign(targets=[node.target], value=node.value), node
                )
            )
        else:
            return None

    def visit_Module(self, node):
        node = self.generic_visit(node)
        node.body = [
            stmt
            for stmt in node.body
            if not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant))
        ] or [ast.copy_location(ast.Pass(), node)]
        return node

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef | ast.Assign:
        node = self.generic_visit(node)
        node.body = [
            stmt
            for stmt in node.body
            if not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant))
        ] or [ast.copy_location(ast.Pass(), node)]

        if len(node.body) == 1 and len(node.bases) == 1:
            return ast.Assign(
                [ast.Name(node.name, ast.Store())],
                node.bases[0],
                lineno=0,
                col_offset=0,
            )
        return node


def pre_optimize(module, name):
    return PreOptimize().visit(module)

def spike_prime_compyne(input_module: Path, slot: int = 0, debug_build: bool = False):
    sys.path.append(str(input_module.parent))
    compyner = ComPYner(
        exclude_modules=SPIKE_PRIME_MODULES,
        module_preprocessor=pre_optimize,
        # pastprocessor=past_optimize,
        require_dunder_name=debug_build,
        random_name_length=0 if debug_build else 4,
        keep_names=debug_build,
    )
    code = f"# LEGO type:standard slot:{slot} autostart\n" + compyner.compyne_from_ast(
        "__main__",
        ast_from_file(input_module),
        origin=input_module.absolute(),
    )
    sys.path.pop()
    return code
